/**
 * Copyright  2005 Imperial College, London, England.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Contact: Kevin Twidle <kpt@doc.ic.ac.uk>
 *
 * Created on Nov 15, 2005
 *
 * $Log: DiscoveryManager.java,v $
 * Revision 1.5  2006/03/15 14:40:46  kpt
 * Kept wrong version, stepped back
 *
 * Revision 1.3  2005/11/23 14:47:07  kpt
 * Removed Iterable from QDParser.
 *
 * Revision 1.2  2005/11/17 11:36:28  kpt
 * Added context to many ops for arg substitution
 *
 * Revision 1.1  2005/11/15 16:33:25  kpt
 * Result added to create
 *
 */

package net.ponder2.managedobject;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import net.ponder2.ManagedObject;
import net.ponder2.apt.Ponder2op;
import net.ponder2.bsn.BSNReceiverInterface;
import net.ponder2.bsn.InsulinPump;
import net.ponder2.exception.Ponder2Exception;
import net.ponder2.objects.P2Object;

/**
 * The Discovery Manager receives heart beats and values from BSN nodes. The
 * Discovery Manager keeps track of BSN nodes, when they arrive, when they
 * disappear, and sends the appropriate event. Whenever a value arrives, it is
 * filtered to see if it has changed and if so, it is sent on as an event.
 * <p>
 * Part of the Ponder2 Tutorial.
 * 
 * @author Kevin Twidle
 * @version $Id:$
 */
public class DiscoveryManager implements ManagedObject {

  Map<String, DeviceRecord> db;
  private P2Object newBSN;
  private P2Object lostBSN;
  private P2Object newPump;
  private P2Object lostPump;
  private P2Object myP2Object;
  private P2Object bsnValue;

  private Map<String, Integer> value;
  private Map<String, Integer> rate;
  private boolean filter;

  /**
   * Creates a new instance of the Discovery Manager. There can only be one
   * active Discovery manager running at a time because it listens for BSN
   * events on the RMI port BSNBus.
   * 
   * @param myP2Object
   */
  @Ponder2op("create:")
  public DiscoveryManager(P2Object myP2Object, String patientName) {
    this.myP2Object = myP2Object;
    value = new HashMap<String, Integer>();
    rate = new HashMap<String, Integer>();
    filter = true;
    try {
      Naming.rebind(patientName + "_" + "BSNBus", new BSNReceiver() {

        @Override
        public void heartbeat(String type, String name, int rate) throws RemoteException {
          try {
            synchronized (value) {
              rxHeartBeat(type, name);
            }
          }
          catch (Ponder2Exception e) {
            System.out.println(e.getMessage());
          }
          catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
        }

        @Override
        public void value(String type, String name, int newValue, int rate) throws RemoteException {
          try {
            synchronized (value) {
              rxValue(type, name, newValue, rate);
            }
          }
          catch (Exception e) {
            System.out.println("Discovery Manager error: " + e.getMessage());
            // e.printStackTrace();
          }
        }
      });
    }
    catch (RemoteException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    catch (MalformedURLException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    db = new HashMap<String, DeviceRecord>();
    startPoll();
  }

  /**
   * Sets the event to be generated when a new Insulin pump is detected
   * 
   * @param anEventType
   */
  @Ponder2op("newPump:")
  public void newPump(P2Object anEventType) {
    this.newPump = anEventType;
  }

  /**
   * Sets the event to be generated when an Insulin pump is lost
   * 
   * @param anEventType
   */
  @Ponder2op("lostPump:")
  public void lostPump(P2Object anEventType) {
    this.lostPump = anEventType;
  }

  /**
   * Sets the event to be generated when a new BSN node is detected
   * 
   * @param anEventType
   */
  @Ponder2op("newBSN:")
  public void newBsnEvent(P2Object anEventType) {
    this.newBSN = anEventType;
  }

  /**
   * Sets the event to be generated when a BSN node is lost
   * 
   * @param anEventType
   */
  @Ponder2op("lostBSN:")
  public void lostBsnEvent(P2Object anEventType) {
    this.lostBSN = anEventType;
  }

  /**
   * Sets the event to be generated when a value is received from a BSN node
   * 
   * @param anEventType
   */
  @Ponder2op("bsnValue:")
  public void bsnValue(P2Object anEventType) {
    this.bsnValue = anEventType;
  }

  /**
   * If aBoolean is true then BSN values are filtered to weed out duplicate
   * values.
   * <P>
   * This can be set using PonderTalk commands e.g.
   * 
   * <pre>
   * discmngr filter: true.
   * </pre>
   * 
   * or
   * 
   * <pre>
   * discmngr filter: false.
   * </pre>
   * 
   * @param filter
   */
  @Ponder2op("filter:")
  public void bsnFilter(boolean aBoolean) {
    filter = aBoolean;
  }

  /**
   * @param type
   * @param name
   * @throws Ponder2Exception
   */
  protected void rxHeartBeat(String type, String name) throws Ponder2Exception {
    // System.out.println("Discovery manager: rxheartbeat from " + type + " " +
    // name);
    if (db.get(name) == null) {
      if (type.equals(InsulinPump.TYPE)) {
        if (newPump != null)
          sendEvent(newPump, P2Object.create(name));
      }
      else {
        if (newBSN != null)
          sendEvent(newBSN, P2Object.create(type), P2Object.create(name));
      }
    }
    db.put(name, new DeviceRecord(name, type, new Date().getTime()));
  }

  /**
   * @param type
   * @param name
   * @param oldValue
   * @param newValue
   * @param rate
   * @throws Ponder2Exception
   */
  protected void rxValue(String type, String name, int newValue, int newRate)
      throws Ponder2Exception {
    int oldValue = -1;
    int oldRate = -1;
    // System.out.println("name " + name + " newValue " + newValue);
    if (value.containsKey(name))
      oldValue = value.get(name);
    if (rate.containsKey(name))
      oldRate = rate.get(name);
    if (oldValue != newValue || oldRate != newRate || !filter) {
      sendEvent(bsnValue, P2Object.create(name), P2Object.create(newValue), P2Object
          .create(oldValue), P2Object.create(newRate));
      value.put(name, newValue);
      rate.put(name, newRate);
    }
  }

  private void sendEvent(P2Object event, P2Object... args) throws Ponder2Exception {
    event.operation(myP2Object, "create:", P2Object.create(args));

  }

  private void startPoll() {
    new Thread() {

      @Override
      public void run() {
        while (true) {
          try {
            long time = new Date().getTime();
            for (Iterator<String> it = db.keySet().iterator(); it.hasNext();) {
              String bsnName = it.next();
              DeviceRecord rec = db.get(bsnName);
              long timeStamp = rec.time;
              long last = (time - timeStamp) / 1000;
              if (last > 6) {
                removeDevice(rec);
                it.remove();
              }
            }
            Thread.sleep(2000);
          }
          catch (InterruptedException e) {
          }
        }
      }
    }.start();
  }

  /**
   * @param remove
   */
  private void removeDevice(DeviceRecord remove) {
    try {
      if (remove.type.equals(InsulinPump.TYPE)) {
        if (lostPump != null)
          sendEvent(lostPump, P2Object.create(remove.name));
      }
      else {
        if (lostBSN != null)
          sendEvent(lostBSN, P2Object.create(remove.name));
      }
    }
    catch (Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  public abstract class BSNReceiver extends UnicastRemoteObject implements BSNReceiverInterface {

    public BSNReceiver() throws RemoteException {
    }

    /*
     * (non-Javadoc)
     * 
     * @see BSNDiscoveryInterface#heartbeat(java.lang.String, java.lang.String)
     */
    public abstract void heartbeat(String type, String name, int rate) throws RemoteException;

    /*
     * (non-Javadoc)
     * 
     * @see BSNDiscoveryInterface#value(java.lang.String, java.lang.String,
     * double)
     */
    public abstract void value(String type, String name, int newValue, int rate)
        throws RemoteException;
  }

  public class DeviceRecord {

    String name;
    String type;
    long time;

    /**
     * @param name
     * @param type
     * @param time
     */
    public DeviceRecord(String name, String type, long time) {
      this.name = name;
      this.type = type;
      this.time = time;
    }
  }

}
