/**
 * Copyright  2008 Kevin Twidle, Imperial College, London, England.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Contact: Kevin Twidle <kpt@doc.ic.ac.uk>
 *
 * Created on Oct 5, 2008
 *
 * $Log:$
 */

package net.ponder2.managedobject;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import net.ponder2.ManagedObject;
import net.ponder2.apt.Ponder2op;
import net.ponder2.exception.Ponder2AuthorizationException;
import net.ponder2.exception.Ponder2Exception;
import net.ponder2.nurse.BsnPanel;
import net.ponder2.nurse.NurseFrame;
import net.ponder2.objects.P2Hash;
import net.ponder2.objects.P2Object;

/**
 * Nurse Managed Object for the Ponder2 Tutorial. When given a bed (domain) to
 * manage, a window is opened, the domain is scanned for Body Sensor Nodes. Any
 * nodes found are added to the window and the nodes are asked for their
 * readings and scan rates. Scan rates can be altered by using the drop-down in
 * the Nurse window.
 * 
 * @author Kevin Twidle
 * @version $Id:$
 */
public class Nurse implements PropertyChangeListener, ManagedObject, Runnable {

  private P2Object myP2Object;
  private NurseFrame window;
  Map<String, BsnPanel> panel = new HashMap<String, BsnPanel>();
  Map<String, P2Object> bsnmo = new HashMap<String, P2Object>();
  boolean running = true;
  private P2Object bed;
  private String name;

  /**
   * Creates a nurse with the given name.
   * 
   * @param name
   *          the Nurse's name
   */
  @Ponder2op("create:")
  public Nurse(P2Object myP2Object, String name) {
    this.myP2Object = myP2Object;
    this.name = name;
    window = new NurseFrame(name);
    window.setVisible(true);
  }

  /**
   * Tells the Nurse to manage a particular bed
   * 
   * @param bed
   * @throws Ponder2Exception
   */
  @Ponder2op("manageBed:")
  public void manageBed(P2Object bed) throws Ponder2Exception {
    if (this.bed != bed) {
      this.bed = bed;
      new Thread(this).start();
    }
  }

  /**
   * Tells the nurse to manage all the beds in a ward - Not implemented yet
   * 
   * @param bed
   */
  @Ponder2op("manageWard:")
  public void manageWard(P2Object wardDomain) {

  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent
   * )
   */
  public void propertyChange(PropertyChangeEvent evt) {
    System.out.println("Property change " + evt.getPropertyName() + " " + evt.getOldValue()
        + " => " + evt.getNewValue() + " " + this + " " + evt);
    BsnPanel bsnPanel = (BsnPanel)evt.getSource();
    P2Object bsn = bsnmo.get(bsnPanel.getBsnName());
    int value = (int)(Double.parseDouble((String)evt.getNewValue()) * 1000.0);
    try {
      bsn.operation(myP2Object, "setSampleRate:", P2Object.create(value));
      window.log("");
    }
    catch (Ponder2AuthorizationException e) {
      System.out.println("Nurse: not authorised");
      bsnPanel.setBsnRate("" + (int)(Double.parseDouble((String)evt.getOldValue()) * 1000.0));
      window.log("Error - Not authorised to set the rate");
    }
    catch (Ponder2Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Runnable#run()
   */
  public void run() {
    while (running) {
      // Get all the known BSN names first
      Set<String> allBsns = new HashSet<String>(panel.keySet());
      try {
        P2Object bsndomain;

        bsndomain = bed.operation(myP2Object, "at:", P2Object.create("bsn"));
        P2Hash list = (P2Hash)bsndomain.operation(myP2Object, "asHash");
        // System.out.println("Nurse: scanning domain");
        Map<String, P2Object> map = list.asMap();
        for (Iterator<String> iterator = map.keySet().iterator(); iterator.hasNext();) {
          String name = iterator.next();
          // System.out.println("Nurse has found " + name);
          P2Object bsn = map.get(name);
          bsnmo.put(name, bsn);
          // Read the rate and value
          double rate = bsn.operation(myP2Object, "getSampleRate").asDouble();
          // System.out.println("Nurse Rate is "+rate);
          String value = bsn.operation(myP2Object, "getValue").asString();
          // System.out.println("Value is "+value);
          BsnPanel bsnPanel = window.addBsn(name, this);
          panel.put(name, bsnPanel);
          bsnPanel.setValue(value);
          bsnPanel.setBsnRate("" + rate);
          allBsns.remove(name);
        }

      }
      catch (Ponder2AuthorizationException e) {
        window.log("Error: cannot access bed");
      }
      catch (Ponder2Exception e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }

      // Remove any missing panels
      for (Iterator<String> it = allBsns.iterator(); it.hasNext();) {
        String name = it.next();
        window.removeBsn(name);
      }

      try {
        Thread.sleep(2000);
      }
      catch (InterruptedException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }
  }

}
